#include "btree_gist.h"

typedef struct intvkey
{
#ifdef HAVE_INT64_TIMESTAMP
   int64       lower,
               upper;
#else
   double      lower,
               upper;
#endif
}   INTVKEY;


/*
** interval ops
*/

PG_FUNCTION_INFO_V1(gintv_compress);
PG_FUNCTION_INFO_V1(gintv_union);
PG_FUNCTION_INFO_V1(gintv_picksplit);
PG_FUNCTION_INFO_V1(gintv_consistent);
PG_FUNCTION_INFO_V1(gintv_penalty);
PG_FUNCTION_INFO_V1(gintv_same);

Datum    gintv_compress(PG_FUNCTION_ARGS);
Datum    gintv_union(PG_FUNCTION_ARGS);
Datum    gintv_picksplit(PG_FUNCTION_ARGS);
Datum    gintv_consistent(PG_FUNCTION_ARGS);
Datum    gintv_penalty(PG_FUNCTION_ARGS);
Datum    gintv_same(PG_FUNCTION_ARGS);

static void gintv_binary_union(Datum *r1, char *r2);
static int  intvkey_cmp(const void *a, const void *b);

/* define for comparison */
#define INTVGE( t1, t2 ) ( t1>=t2 )
#define INTVGT( t1, t2 ) ( t1>t2 )
#define INTVEQ( t1, t2 ) ( t1==t2 )
#define INTVLT( t1, t2 ) ( t1<t2 )
#define INTVLE( t1, t2 ) ( t1<=t2 )





/**************************************************
 * interval ops
 **************************************************/


static void gintv_key_gen ( INTVKEY * k , Interval * i ){

  k->lower = i->time ;
  if (i->month != 0)
#ifdef HAVE_INT64_TIMESTAMP
      k->lower += ((i->month * INT64CONST(30) * INT64CONST(86400000000)));
#else
      k->lower += (i->month * (30.0 * 86400));
#endif
  k->upper = k->lower;

} 



Datum
gintv_compress(PG_FUNCTION_ARGS)
{
  GISTENTRY  *entry = (GISTENTRY *) PG_GETARG_POINTER(0);
  GISTENTRY  *retval;

  if (entry->leafkey)
  {
    INTVKEY     *r = (INTVKEY *) palloc(sizeof(INTVKEY));
    retval = palloc(sizeof(GISTENTRY));
    gintv_key_gen ( r , ( Interval * ) DatumGetPointer (entry->key) );
    gistentryinit(*retval, PointerGetDatum(r),
            entry->rel, entry->page,
            entry->offset, sizeof(INTVKEY), FALSE);
  }
  else
    retval = entry;
  PG_RETURN_POINTER(retval);
}




Datum
gintv_consistent(PG_FUNCTION_ARGS)
{
  GISTENTRY  *entry  = (GISTENTRY *) PG_GETARG_POINTER(0);
  Interval   *query  =  PG_GETARG_INTERVAL_P ( 1 ) ;
  StrategyNumber strategy = (StrategyNumber) PG_GETARG_UINT16(2);

  bool        retval ;
  INTVKEY     *key   ;
  INTVKEY      qqq   ;

  if (!entry->key)  return FALSE;
  key = (INTVKEY *) DatumGetPointer(entry->key);

  gintv_key_gen ( &qqq , query );

  switch (strategy)
  {
    case BTLessEqualStrategyNumber:
      retval = INTVGE( qqq.lower, key->lower );
      break;
    case BTLessStrategyNumber:
      if (GIST_LEAF(entry))
        retval = INTVGT(qqq.lower, key->lower);
      else
        retval = INTVGE(qqq.lower, key->lower);
      break;
    case BTEqualStrategyNumber:
      // in leaf page key->lower always = key->upper
      if (GIST_LEAF(entry))
        retval = INTVEQ(qqq.lower, key->lower);
      else
        retval = (INTVLE( key->lower, qqq.lower) && INTVLE(qqq.lower, key->upper));
      break;
    case BTGreaterStrategyNumber:
      if (GIST_LEAF(entry))
        retval = INTVLT(qqq.lower, key->upper);
      else
        retval = INTVLE(qqq.lower, key->upper);
      break;
    case BTGreaterEqualStrategyNumber:
      retval = INTVLE( qqq.lower, key->upper );
      break;
    default:
      retval = FALSE;
  }

  PG_RETURN_BOOL( retval );
}



Datum
gintv_union(PG_FUNCTION_ARGS)
{
  bytea     *entryvec = (bytea *) PG_GETARG_POINTER(0);
  int               i ,
            numranges ;
  INTVKEY        *cur ,
                 *out = palloc(sizeof(INTVKEY));

  numranges = (VARSIZE(entryvec) - VARHDRSZ) / sizeof(GISTENTRY);
  *(int *) PG_GETARG_POINTER(1) = sizeof(INTVKEY);

  cur = (INTVKEY *) DatumGetPointer((((GISTENTRY *) (VARDATA(entryvec)))[0].key));
  out->lower = cur->lower;
  out->upper = cur->upper;

  for (i = 1; i < numranges; i++)
  {
    cur = (INTVKEY *) DatumGetPointer((((GISTENTRY *) (VARDATA(entryvec)))[i].key));
    if (INTVGT(out->lower, cur->lower))
      out->lower = cur->lower;
    if (INTVLT(out->upper, cur->upper))
      out->upper = cur->upper;
  }

  PG_RETURN_POINTER(out);
}


Datum
gintv_penalty(PG_FUNCTION_ARGS)   
{
	INTVKEY    *origentry = (INTVKEY *) DatumGetPointer(((GISTENTRY *) PG_GETARG_POINTER(0))->key);
        INTVKEY    *newentry  = (INTVKEY *) DatumGetPointer(((GISTENTRY *) PG_GETARG_POINTER(1))->key);
        float      *result = (float *) PG_GETARG_POINTER(2);

        *result = Max(newentry->upper - origentry->upper, 0) +
                Max(origentry->lower - newentry->lower, 0);

        PG_RETURN_POINTER(result);
}


Datum
gintv_picksplit(PG_FUNCTION_ARGS)
{
  PG_RETURN_POINTER(btree_picksplit(
                  (bytea *) PG_GETARG_POINTER(0),
                  (GIST_SPLITVEC *) PG_GETARG_POINTER(1),
                  gintv_binary_union,
                  intvkey_cmp
                ));
}

Datum
gintv_same(PG_FUNCTION_ARGS)
{
  INTVKEY     *b1 = (INTVKEY *) PG_GETARG_POINTER(0);
  INTVKEY     *b2 = (INTVKEY *) PG_GETARG_POINTER(1);

  bool     *result = (bool *) PG_GETARG_POINTER(2);

  if (b1 && b2)
    *result = (INTVEQ(b1->lower, b2->lower) && INTVEQ(b1->upper, b2->upper)) ? TRUE : FALSE;
  else
    *result = (b1 == NULL && b2 == NULL) ? TRUE : FALSE;
  PG_RETURN_POINTER(result);
}

static void
gintv_binary_union(Datum *r1, char *r2)
{
  INTVKEY     *b1;
  INTVKEY     *b2 = (INTVKEY *) r2;

  if (!DatumGetPointer(*r1))
  {
    *r1 = PointerGetDatum(palloc(sizeof(INTVKEY)));
    b1 = (INTVKEY *) DatumGetPointer(*r1);
    b1->upper = b2->upper;
    b1->lower = b2->lower;
  }
  else
  {
    b1 = (INTVKEY *) DatumGetPointer(*r1);

    b1->lower = (INTVGT(b1->lower, b2->lower)) ?
      b2->lower : b1->lower;
    b1->upper = (INTVGT(b1->upper,b2->upper)) ?
      b1->upper : b2->upper;
  }
}


static int
intvkey_cmp(const void *a, const void *b)
{

    if (((INTVKEY *) (((RIX *) a)->r))->lower > ((INTVKEY *) (((RIX *) b)->r))->lower){
      return 1;
    } else if (((INTVKEY *) (((RIX *) a)->r))->lower < ((INTVKEY *) (((RIX *) b)->r))->lower){
      return -1;
    } else {
      return 0;
    }

}

